# -*- Mode: Python; tab-width: 4 -*-
#	Author: Sam Rushing <rushing@nightmare.com>
#
# "Freeze" a Python script into a binary.
#
# This win32 version is based on the original

# TODO: handle 'pycbox' files.

# Import standard modules

import cmp
import os
import string
import sys
import getopt

# Import the freeze-private modules

import findmodules
import makefreeze

# Main program

def main ():
	frozen_c	= './build/frozen.c'
	makefile	= './build/Makefile'
	frozenmain	= './build/frozenmain.cpp'

	resources = []
	subsystem = 'console'

	box_list = []
	pfl_list = []
	modules = []

	if len(sys.argv) < 4:
		print '\nUsage: %s <python_src_dir> <main.py> "console" | "windows" [files...]' % (sys.argv[0])
		print '  where <files> can be additional python modules, windows resource files,'
		print '  python frozen libraries (.pfl files), or pycbox (.db) libraries'
		sys.exit(0)

	python_src = sys.argv[1]
	if not os.path.isfile (os.path.join (python_src, 'Include/Python.h')):
		print '"%s" does not appear to be a Python source directory' % python_src
		sys.exit(0)

	if len(sys.argv) > 3:

		subsystem = sys.argv[3]
		if subsystem not in ['windows', 'console']:
			print "<subsystem> must be one of 'windows' or 'console'"
			sys.exit(0)

		if len (sys.argv) > 4:
			for arg in sys.argv[4:]:
				ext = string.lower (os.path.splitext(arg)[1])
				if not os.path.isabs (arg):
					arg = '../' + arg 
				if ext == '.res':
					resources.append (arg)
				elif ext == '.pfl':
					pfl_list.append (arg)
				elif ext == '.db':
					box_list.append (arg)
				elif ext == '.py':
					modules.append (arg)
				else:
					print 'Hey, what is "%s"?? - ignoring...' % arg

	scriptfile = sys.argv[2]

	skip_modules = {}

	if pfl_list:
		import pfl
		for mod in pfl_list:
			d = pfl.get_pfl_directory (mod)
			for k in d.keys():
				skip_modules[k] = mod

	if box_list:
		import bsddb
		for mod in box_list:
			d = bsddb.btopen (mod, 'rb')
			for k in d.keys():
				skip_modules[k] = mod

	# derive target name from script name
	base = os.path.basename(scriptfile)
	base, ext = os.path.splitext(base)
	if base:
		if base != scriptfile:
			target = base
		else:
			target = base + '.bin'

	# Actual work starts here...

	dict = findmodules.findmodules(scriptfile, modules, sys.path)
	names = dict.keys()
	names.sort()

	if 'os' in names:
		# nother special-case hack
		dict['ntpath'] = findmodules.findmodule ('ntpath', sys.path)
		names.append ('ntpath')

	# remove any modules referenced in a pfl or box

	if pfl_list or box_list:
		# TODO: check for duplicated modules
		for key in dict.keys():
			if skip_modules.has_key (key):
				del dict[key]

	print "Modules being frozen:"
	for name in names:
	    print '\t', name

	backup = frozen_c + '~'
	try:
		os.rename(frozen_c, backup)
	except os.error:
		backup = None

	if not os.path.isdir ('./build'):
		os.mkdir ('./build', 755)

	outfp = open(frozen_c, 'w')
	try:
		makefreeze.makefreeze(outfp, dict)
	finally:
		outfp.close()
	if backup:
		if cmp.cmp(backup, frozen_c):
			sys.stderr.write('%s not changed, not written\n' %
					 frozen_c)
			os.rename(backup, frozen_c)

	builtins = []
	unknown = []
	mods = dict.keys()
	mods.sort()
	for mod in mods:
		if dict[mod] == '<builtin>':
			builtins.append(mod)
		elif dict[mod] == '<unknown>':
			unknown.append(mod)

	addfiles = []

	pyd_modules = match_pyd (unknown)
	if pyd_modules:
		print 'Your app will require the following pyd modules: '
		for mod in pyd_modules:
			print mod

	if pfl_list:
		print 'Your app will require the following python frozen libraries (pfl files):'
		for pfl in pfl_list:
			print pfl

	if box_list:
		print 'Your app will require the following pycbox libraries (db files):'
		for box in box_list:
			print box

	for pyd in pyd_modules:
		unknown.remove (pyd)

	for x in ['posix', 'pwd']:
		if x in unknown:
			unknown.remove (x)

	if unknown:
		sys.stderr.write('Warning: unknown modules remain: %s\n' %
				 string.join(unknown))

	args = string.join (sys.argv)

	# write the makefile
	open (makefile, 'w').write (
		makefile_template % (
			python_src,
			target,
			subsystem,
			string.join (resources)
			)
		)

	pfl_load_libs = []
	# build tag list for pfl's
	for pfl in pfl_list:
		base, file = os.path.split (pfl)
		mod, ext = os.path.splitext (file)
		pfl_load_libs.append ('load_python_frozen_library ("%s.pfl");' % mod)

	if box_list and 'pycbox' not in pfl_list:
		pfl_load_libs.append ('load_python_frozen_library ("pycbox.pfl");')
		pfl_load_libs.append ('PyImport_ImportFrozenModule ("pycbox");')

	open (frozenmain, 'w').write (
		frozenmain_cpp % (
			string.joinfields (pfl_load_libs, '\n')
			)
		)

	print 'building executable...'
	os.chdir ('./build')
	os.system ('nmake')
	#print "\nNow type 'nmake' to build the executable"

def find_pyd ():
	pyd_files = {}
	for path in sys.path:
		if os.path.isdir (path):
			for file in os.listdir (path):
				name = string.lower (file)
				if name[-4:] == '.pyd':
					pyd_files[name] = 1
	return pyd_files

def match_pyd (modules):
	pyd_files = find_pyd()
	result = []
	for mod in modules:
		if pyd_files.has_key (string.lower(mod) + '.pyd'):
			result.append (mod)
	return result

makefile_template = \
"""
pythondir = %s
appname   = %s
subsystem = %s
resources = %s

NODEBUG = 1
BUILD_DL = 0

!include <$(pythondir)/make_nt.in>

PYTHON_LIBS= $(pythondir)/vc40/Python$(pyversion).lib

all:	$(appname).exe

frozenmain.obj: $*.cpp frozen.c
	$(CC) $(cdl) -I$(pythondir)/Include -I$(pythondir)/pc $(cflags) $(cdebug) $(cinclude) $(pythonopts) $*.cpp

$(appname).exe : frozenmain.obj
	link -out:$*.exe frozenmain.obj $(PYTHON_LIBS) $(lcustom) $(guilibsdll) shell32.lib comdlg32.lib wsock32.lib -subsystem:$(subsystem) $(resources)
"""

frozenmain_cpp = """
#include <Python.h>
#include "frozen.c"

extern "C" {
int PyImport_ImportFrozenModule (char *);
void PySys_SetArgv (int argc, char ** argv);
extern __declspec(dllimport) struct _frozen * PyImport_FrozenModules;
}

static int
glue_frozen_modules (void)
{
  struct _frozen * p, *q;
  struct _frozen * new_list;
  int nmods = 0;

  // how many modules in the existing list?
  for (p = PyImport_FrozenModules; p->name; p++) {
	nmods++;
  }
  // how many in the new list?
  for (p = _frozen_modules; p->name; p++) {
	nmods++;
  }
  new_list = (struct _frozen *) malloc (sizeof (struct _frozen) * (nmods + 1));
  if (!new_list) {
	return -1;
  }
  q = new_list;
  // copy the old list in
  for (p = PyImport_FrozenModules; p->name; p++, q++) {
	q->name = p->name;
	q->code = p->code;
	q->size = p->size;
  }
  // copy the new list in
  for (p = _frozen_modules; p->name; p++, q++) {
	q->name = p->name;
	q->code = p->code;
	q->size = p->size;
  }
  q->name = 0;
  q->code = 0;
  q->size = 0;
  // plug in the new list
  PyImport_FrozenModules = new_list;
  return 0;
}

#define WIN32_LEAN_AND_MEAN
#include <windows.h>

void
load_python_frozen_library (char * name)
{
  if (!LoadLibrary(name)) {
    char msg[511];
	sprintf (msg, "couldn't load library '%%s'", name);
    MessageBox (0, msg, "Error Loading Library", MB_OK|MB_ICONEXCLAMATION);
	exit(0);
  }
}

// undo rename2.h magic
#undef main
int
main (int argc, char * argv[])
{
  int n, sts;
  Py_Initialize();
  PySys_SetArgv (argc, argv);
  
  %s

  glue_frozen_modules();

  n = PyImport_ImportFrozenModule ("__main__");
  if (n == 0) {
    Py_FatalError ("__main__ not frozen");
  } else if (n < 0) {
    PyErr_Print();
    sts = 1;
  } else {
    sts = 0;
  }
  Py_Exit (sts);
  return sts;
}

#include <windows.h>
int APIENTRY
WinMain (HINSTANCE hInstance,
	 HINSTANCE hPrevInstance,
	 LPSTR lpCmdLine,
	 int nCmdShow)
{
  return main (1, &lpCmdLine);
}  
"""

if __name__ == '__main__':
	main()
